@@ -6,9 +6,9 @@ set -o xtrace
 # Running a unity build (sometimes called a jumbo build) is both a useful thing
 # to test and reduces compile time.
 #
-# Enable all auto-features to ensure that we're proprely testing all optional
+# Enable all auto-features to ensure that we're properly testing all optional
 # dependencies.
-meson build --unity=on --auto-features=enabled $CONFIGURE_OPTIONS
+meson build -Dwerror=true --unity=on --auto-features=enabled $CONFIGURE_OPTIONS
 ninja -C build
 
 xvfb-run --auto-servernum ninja -C build test
