@@ -14,10 +14,6 @@
  * Any additions, deletions, or changes to the original source files
  * must be clearly indicated in accompanying documentation.
  *
- * If only executable code is distributed, then the accompanying
- * documentation must state that "this software is based in part on the
- * work of the Khronos Group."
- *
  * THE MATERIALS ARE PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
  * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
@@ -95,6 +91,8 @@ static DEFINE_INITIALIZED_LKDHASH(__GLXvendorNameHash, __glXVendorNameHash);
 
 typedef struct __GLXdisplayInfoHashRec {
     __GLXdisplayInfo info;
+    Bool inTeardown;
+    XExtCodes *extCodes;
     UT_hash_handle hh;
 } __GLXdisplayInfoHash;
 
@@ -687,6 +685,10 @@ static void CleanupDisplayInfoEntry(void *unused, __GLXdisplayInfoHash *pEntry)
         free(pEntry->info.clientStrings[i]);
     }
 
+    if (pEntry->extCodes != NULL) {
+        XESetCloseDisplay(pEntry->info.dpy, pEntry->extCodes->extension, NULL);
+    }
+
     LKDHASH_TEARDOWN(__GLXvendorXIDMappingHash,
                      pEntry->info.xidVendorHash, NULL, NULL, False);
 }
@@ -697,15 +699,29 @@ static int OnDisplayClosed(Display *dpy, XExtCodes *codes)
 
     LKDHASH_WRLOCK(__glXDisplayInfoHash);
 
+    // Set the inTeardown flag for this display, but don't remove it from the
+    // hashtable yet. We shouldn't try to look up this display after this, but
+    // just in case something does, we don't want to try to set up a new
+    // __GLXdisplayInfoHash struct for it.
     HASH_FIND_PTR(_LH(__glXDisplayInfoHash), &dpy, pEntry);
     if (pEntry != NULL) {
-        __glXDisplayClosed(&pEntry->info);
-        HASH_DEL(_LH(__glXDisplayInfoHash), pEntry);
+        assert(!pEntry->inTeardown);
+        pEntry->inTeardown = True;
     }
     LKDHASH_UNLOCK(__glXDisplayInfoHash);
 
-    CleanupDisplayInfoEntry(NULL, pEntry);
-    free(pEntry);
+    if (pEntry != NULL) {
+        __glXDisplayClosed(&pEntry->info);
+
+        // Now, we can remove the display from the hashtable and free it.
+        LKDHASH_WRLOCK(__glXDisplayInfoHash);
+        HASH_DEL(_LH(__glXDisplayInfoHash), pEntry);
+        LKDHASH_UNLOCK(__glXDisplayInfoHash);
+
+        pEntry->extCodes = NULL;
+        CleanupDisplayInfoEntry(NULL, pEntry);
+        free(pEntry);
+    }
 
     return 0;
 }
@@ -724,6 +740,12 @@ __GLXdisplayInfo *__glXLookupDisplay(Display *dpy)
     LKDHASH_UNLOCK(__glXDisplayInfoHash);
 
     if (pEntry != NULL) {
+        if (pEntry->inTeardown) {
+            // This shouldn't happen: If we're in the process of tearing down a
+            // display, then we shouldn't be trying to look up the display
+            // again.
+            return NULL;
+        }
         return &pEntry->info;
     }
 
@@ -738,15 +760,15 @@ __GLXdisplayInfo *__glXLookupDisplay(Display *dpy)
     LKDHASH_WRLOCK(__glXDisplayInfoHash);
     HASH_FIND_PTR(_LH(__glXDisplayInfoHash), &dpy, foundEntry);
     if (foundEntry == NULL) {
-        XExtCodes *extCodes = XAddExtension(dpy);
-        if (extCodes == NULL) {
+        pEntry->extCodes = XAddExtension(dpy);
+        if (pEntry->extCodes == NULL) {
             CleanupDisplayInfoEntry(NULL, pEntry);
             free(pEntry);
             LKDHASH_UNLOCK(__glXDisplayInfoHash);
             return NULL;
         }
 
-        XESetCloseDisplay(dpy, extCodes->extension, OnDisplayClosed);
+        XESetCloseDisplay(dpy, pEntry->extCodes->extension, OnDisplayClosed);
         HASH_ADD_PTR(_LH(__glXDisplayInfoHash), info.dpy, pEntry);
     } else {
         // Another thread already created the hashtable entry.
